from __future__ import annotations

import argparse
import ast
import shutil
import subprocess
from pathlib import Path
from typing import Any

from django.apps import apps
from django.core.management import BaseCommand, CommandError
from django.db import DatabaseError, connections
from django.db.migrations.recorder import MigrationRecorder

from django_linear_migrations.apps import MigrationDetails, is_first_party_app_config


class Command(BaseCommand):
    help = (
        "Fix a conflict in your migration history by rebasing the conflicting"
        + " migration on to the end of the app's migration history."
    )

    # Checks disabled because the django-linear-migrations' checks would
    # prevent us continuing
    requires_system_checks: list[str] = []

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "app_label",
            help="Specify the app label to rebase the migration for.",
        )

    def handle(self, *args: Any, app_label: str, **options: Any) -> None:
        app_config = apps.get_app_config(app_label)
        if not is_first_party_app_config(app_config):
            raise CommandError(f"{app_label!r} is not a first-party app.")

        migration_details = MigrationDetails(app_label)
        max_migration_txt = migration_details.dir / "max_migration.txt"
        if not max_migration_txt.exists():
            raise CommandError(f"{app_label} does not have a max_migration.txt.")

        migration_names = find_migration_names(
            max_migration_txt.read_text().splitlines()
        )
        if migration_names is None:
            raise CommandError(
                f"{app_label}'s max_migration.txt does not seem to contain a"
                + " merge conflict."
            )
        merged_migration_name, rebased_migration_name = migration_names
        if merged_migration_name not in migration_details.names:
            raise CommandError(
                f"Parsed {merged_migration_name!r} as the already-merged"
                + f" migration name from {app_label}'s max_migration.txt, but"
                + " this migration does not exist."
            )
        if rebased_migration_name not in migration_details.names:
            raise CommandError(
                f"Parsed {rebased_migration_name!r} as the rebased migration"
                + f" name from {app_label}'s max_migration.txt, but this"
                + " migration does not exist."
            )

        rebased_migration_filename = f"{rebased_migration_name}.py"
        rebased_migration_path = migration_details.dir / rebased_migration_filename
        if not rebased_migration_path.exists():
            raise CommandError(
                f"Detected {rebased_migration_filename!r} as the rebased"
                + " migration filename, but it does not exist."
            )

        if migration_applied(app_label, rebased_migration_name):
            raise CommandError(
                f"Detected {rebased_migration_name} as the rebased migration,"
                + " but it is applied to the local database. Undo the rebase,"
                + " reverse the migration, and try again."
            )

        content = rebased_migration_path.read_text()

        try:
            module_def = ast.parse(content)
        except SyntaxError:
            raise CommandError(
                f"Encountered a SyntaxError trying to parse {rebased_migration_filename!r}."
            )

        # Find the migration class
        class_defs = [
            node
            for node in module_def.body
            if isinstance(node, ast.ClassDef) and node.name == "Migration"
        ]
        if not class_defs:
            raise CommandError(
                f"Could not find a Migration class in {rebased_migration_filename!r}."
            )
        if len(class_defs) > 1:
            raise CommandError(
                f"Found multiple Migration classes in {rebased_migration_filename!r}."
            )
        migration_class_def = class_defs[0]

        dependencies_assignments = [
            node
            for node in migration_class_def.body
            if isinstance(node, ast.Assign)
            and len(node.targets) == 1
            and isinstance(node.targets[0], ast.Name)
            and node.targets[0].id == "dependencies"
            and isinstance(node.value, (ast.List, ast.Tuple))
        ]
        if not dependencies_assignments:
            raise CommandError(
                f"Could not find a dependencies = [...] assignment in {rebased_migration_filename!r}."
            )
        if len(dependencies_assignments) > 1:
            raise CommandError(
                f"Found multiple dependencies = [...] assignments in {rebased_migration_filename!r}."
            )

        dependencies = dependencies_assignments[0].value
        assert isinstance(dependencies, (ast.List, ast.Tuple))

        lines = content.splitlines(keepends=True)
        before_deps_len = (
            sum(len(line) for line in lines[: dependencies.lineno - 1])
            + dependencies.col_offset
        )
        assert dependencies.end_lineno is not None
        assert dependencies.end_col_offset is not None
        after_deps_len = (
            sum(len(line) for line in lines[: dependencies.end_lineno - 1])
            + dependencies.end_col_offset
        )

        before_deps = content[:before_deps_len]
        after_deps = content[after_deps_len:]

        if isinstance(dependencies, ast.Tuple):
            new_dependencies: ast.Tuple | ast.List = ast.Tuple(elts=[])
        else:
            new_dependencies = ast.List(elts=[])
        num_this_app_dependencies = 0
        for dependency in dependencies.elts:
            # Skip swappable_dependency calls, other dynamically defined
            # dependencies, and bad definitions
            if (
                not isinstance(dependency, (ast.Tuple, ast.List))
                or len(dependency.elts) != 2
                or not all(
                    isinstance(el, ast.Constant) and isinstance(el.value, str)
                    for el in dependency.elts
                )
            ):
                new_dependencies.elts.append(dependency)
                continue

            dependency_app_label_node = dependency.elts[0]
            assert isinstance(dependency_app_label_node, ast.Constant)
            dependency_app_label = dependency_app_label_node.value
            assert isinstance(dependency_app_label, str)

            if dependency_app_label == app_label:
                num_this_app_dependencies += 1
                new_dependencies.elts.append(
                    ast.Tuple(
                        elts=[
                            ast.Constant(app_label),
                            ast.Constant(merged_migration_name),
                        ]
                    )
                )
            else:
                new_dependencies.elts.append(dependency)

        if num_this_app_dependencies != 1:
            raise CommandError(
                f"Cannot edit {rebased_migration_filename!r} since it has "
                + f"{num_this_app_dependencies} dependencies within "
                + f"{app_label}."
            )

        new_content = before_deps + ast.unparse(new_dependencies) + after_deps

        merged_number, _merged_rest = merged_migration_name.split("_", 1)
        _rebased_number, rebased_rest = rebased_migration_name.split("_", 1)
        new_number = int(merged_number) + 1
        new_name = str(new_number).zfill(4) + "_" + rebased_rest
        new_path_parts = rebased_migration_path.parts[:-1] + (f"{new_name}.py",)
        new_path = Path(*new_path_parts)

        rebased_migration_path.rename(new_path)
        new_path.write_text(new_content)
        max_migration_txt.write_text(f"{new_name}\n")

        black_path = shutil.which("black")
        if black_path:  # pragma: no cover
            subprocess.run(
                [black_path, "--fast", "--", new_path],
                capture_output=True,
            )

        self.stdout.write(
            f"Renamed {rebased_migration_path.parts[-1]} to {new_path.parts[-1]},"
            + " updated its dependencies, and updated max_migration.txt."
        )


def find_migration_names(max_migration_lines: list[str]) -> tuple[str, str] | None:
    lines = max_migration_lines
    if len(lines) <= 1:
        return None
    if not lines[0].startswith("<<<<<<<"):
        return None
    if not lines[-1].startswith(">>>>>>>"):
        return None
    migration_names = (lines[1].strip(), lines[-2].strip())
    if is_merge_in_progress():
        # During the merge 'ours' and 'theirs' are swapped in comparison with rebase
        migration_names = (migration_names[1], migration_names[0])
    return migration_names


def is_merge_in_progress() -> bool:
    try:
        subprocess.run(
            ["git", "rev-parse", "--verify", "MERGE_HEAD"],
            capture_output=True,
            check=True,
            text=True,
        )
    except (FileNotFoundError, subprocess.SubprocessError):
        # Either:
        # - `git` is not available, or broken
        # - there is no git repository
        # - no merge head exists, so assume rebasing
        return False
    # Merged head exists, we are merging
    return True


def migration_applied(app_label: str, migration_name: str) -> bool:
    Migration = MigrationRecorder.Migration
    for alias in connections:
        try:
            if (
                Migration.objects.using(alias)
                .filter(app=app_label, name=migration_name)
                .exists()
            ):
                return True
        except DatabaseError:
            # django_migrations table does not exist -> no migrations applied
            pass
    return False



from __future__ import annotations

import argparse
import sys
from typing import Any

from django.apps import apps
from django.core.management.commands.makemigrations import Command as BaseCommand
from django.db.migrations.loader import MigrationLoader

from django_linear_migrations.apps import (
    MigrationDetails,
    first_party_app_configs,
    get_graph_plan,
)


class Command(BaseCommand):
    help = "Generate max_migration.txt files for first-party apps."

    # Checks disabled because the django-linear-migrations' checks would
    # prevent us continuing
    requires_system_checks: list[str] = []

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "args",
            metavar="app_label",
            nargs="*",
            help="Specify the app label(s) to create max migration files for.",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            default=False,
            help="Actually create the files.",
        )
        parser.add_argument(
            "--recreate",
            action="store_true",
            default=False,
            help=(
                "Recreate existing files. By default only non-existing files"
                + " will be created."
            ),
        )

    def handle(
        self, *app_labels: str, dry_run: bool, recreate: bool, **options: Any
    ) -> None:
        # Copied check from makemigrations
        labels = set(app_labels)
        has_bad_labels = False
        for app_label in labels:
            try:
                apps.get_app_config(app_label)
            except LookupError as err:
                self.stderr.write(str(err))
                has_bad_labels = True
        if has_bad_labels:
            sys.exit(2)

        any_created = False
        migration_loader = MigrationLoader(None, ignore_no_migrations=True)
        graph_plan = get_graph_plan(loader=migration_loader, app_labels=labels)
        for app_config in first_party_app_configs():
            if labels and app_config.label not in labels:
                continue

            migration_details = MigrationDetails(app_config.label)
            if not migration_details.has_migrations:
                continue

            max_migration_txt = migration_details.dir / "max_migration.txt"
            if recreate or not max_migration_txt.exists():
                if not dry_run:
                    max_migration_name = [
                        k[1] for k in graph_plan if k[0] == app_config.label
                    ][-1]
                    max_migration_txt.write_text(max_migration_name + "\n")
                    self.stdout.write(
                        f"Created max_migration.txt for {app_config.label}."
                    )
                else:
                    self.stdout.write(
                        f"Would create max_migration.txt for {app_config.label}."
                    )
                any_created = True

        if not any_created:
            self.stdout.write("No max_migration.txt files need creating.")



from __future__ import annotations

import pkgutil
from collections.abc import Generator, Iterable
from functools import lru_cache
from importlib import import_module, reload
from pathlib import Path
from types import ModuleType

from django.apps import AppConfig, apps
from django.conf import settings
from django.core.checks import Error, Tags, register
from django.core.signals import setting_changed
from django.db.migrations.loader import MigrationLoader
from django.dispatch import receiver
from django.utils.functional import cached_property


class DjangoLinearMigrationsAppConfig(AppConfig):
    name = "django_linear_migrations"
    verbose_name = "django-linear-migrations"

    def ready(self) -> None:
        register(Tags.models)(check_max_migration_files)


@lru_cache(maxsize=1)
def get_first_party_app_labels() -> set[str] | None:
    if not settings.is_overridden("FIRST_PARTY_APPS"):
        return None
    return {AppConfig.create(name).label for name in settings.FIRST_PARTY_APPS}


@receiver(setting_changed)
def reset_first_party_app_labels(*, setting: str, **kwargs: object) -> None:
    if setting == "FIRST_PARTY_APPS":
        get_first_party_app_labels.cache_clear()


def is_first_party_app_config(app_config: AppConfig) -> bool:
    first_party_labels = get_first_party_app_labels()
    if first_party_labels is not None:
        return app_config.label in first_party_labels

    # Check if it seems to be installed in a virtualenv
    path = Path(app_config.path)
    return "site-packages" not in path.parts and "dist-packages" not in path.parts


def first_party_app_configs() -> Generator[AppConfig]:
    for app_config in apps.get_app_configs():
        if is_first_party_app_config(app_config):
            yield app_config


class MigrationDetails:
    migrations_module_name: str | None
    migrations_module: ModuleType | None

    def __init__(self, app_label: str, do_reload: bool = False) -> None:
        self.app_label = app_label

        # Some logic duplicated from MigrationLoader.load_disk, but avoiding
        # loading all migrations since that's relatively slow.
        (
            self.migrations_module_name,
            _explicit,
        ) = MigrationLoader.migrations_module(app_label)
        if self.migrations_module_name is None:
            self.migrations_module = None
        else:
            try:
                self.migrations_module = import_module(self.migrations_module_name)
            except ModuleNotFoundError:
                # Unmigrated app
                self.migrations_module = None
            else:
                if do_reload:
                    reload(self.migrations_module)

    @property
    def has_migrations(self) -> bool:
        return (
            self.migrations_module is not None
            # Not namespace module:
            and self.migrations_module.__file__ is not None
            # Django ignores non-package migrations modules
            and hasattr(self.migrations_module, "__path__")
            and len(self.names) > 0
        )

    @cached_property
    def dir(self) -> Path:
        assert self.migrations_module is not None
        module_file = self.migrations_module.__file__
        assert module_file is not None
        return Path(module_file).parent

    @cached_property
    def names(self) -> set[str]:
        assert self.migrations_module is not None
        path = self.migrations_module.__path__
        return {
            name
            for _, name, is_pkg in pkgutil.iter_modules(path)
            if not is_pkg and name[0] not in "_~"
        }


def get_graph_plan(
    loader: MigrationLoader, app_labels: Iterable[str] | None = None
) -> list[tuple[str, str]]:
    nodes = loader.graph.leaf_nodes()
    if app_labels:
        nodes = [
            (app_label, name)
            for app_label, name in loader.graph.leaf_nodes()
            if app_label in app_labels
        ]
    plan: list[tuple[str, str]] = loader.graph._generate_plan(nodes, at_end=True)
    return plan


def check_max_migration_files(
    *, app_configs: Iterable[AppConfig] | None = None, **kwargs: object
) -> list[Error]:
    errors = []
    if app_configs is not None:
        app_config_set = set(app_configs)
    else:
        app_config_set = set()

    migration_loader = MigrationLoader(None, ignore_no_migrations=True)
    app_labels = [a.label for a in first_party_app_configs()]
    conflicts = {
        app_label: names
        for app_label, names in migration_loader.detect_conflicts().items()
        if app_label in app_labels
    }
    if conflicts:
        conflict_msg = "".join(
            f"\n* {app_label}: {', '.join(sorted(names))}"
            for app_label, names in conflicts.items()
        )
        errors.append(
            Error(
                id="dlm.E005",
                msg=(
                    "Conflicting migrations detected - multiple leaf nodes "
                    + f"detected for these apps:{conflict_msg}"
                ),
                hint=(
                    "Fix the conflict, e.g. with "
                    + "'./manage.py makemigrations --merge --skip-checks'."
                ),
            )
        )
        return errors

    graph_plan = get_graph_plan(loader=migration_loader, app_labels=app_labels)
    for app_config in first_party_app_configs():
        # When only checking certain apps, skip the others
        if app_configs is not None and app_config not in app_config_set:
            continue
        app_label = app_config.label
        migration_details = MigrationDetails(app_label)

        if not migration_details.has_migrations:
            continue

        max_migration_txt = migration_details.dir / "max_migration.txt"
        if not max_migration_txt.exists():
            errors.append(
                Error(
                    id="dlm.E001",
                    msg=f"{app_label}'s max_migration.txt does not exist.",
                    hint=(
                        "If you just installed django-linear-migrations, run"
                        + " 'python manage.py create_max_migration_files'."
                        + " Otherwise, check how it has gone missing."
                    ),
                )
            )
            continue

        max_migration_txt_lines = max_migration_txt.read_text().strip().splitlines()
        if len(max_migration_txt_lines) > 1:
            errors.append(
                Error(
                    id="dlm.E002",
                    msg=f"{app_label}'s max_migration.txt contains multiple lines.",
                    hint=(
                        "This may be the result of a git merge. Fix the file"
                        + " to contain only the name of the latest migration,"
                        + " or maybe use the 'rebase-migration' command."
                    ),
                )
            )
            continue

        max_migration_name = max_migration_txt_lines[0]
        if max_migration_name not in migration_details.names:
            errors.append(
                Error(
                    id="dlm.E003",
                    msg=(
                        f"{app_label}'s max_migration.txt points to"
                        + f" non-existent migration {max_migration_name!r}."
                    ),
                    hint=(
                        "Edit the max_migration.txt to contain the latest"
                        + " migration's name."
                    ),
                )
            )
            continue

        real_max_migration_name = [
            name for gp_app_label, name in graph_plan if gp_app_label == app_label
        ][-1]
        if max_migration_name != real_max_migration_name:
            errors.append(
                Error(
                    id="dlm.E004",
                    msg=(
                        f"{app_label}'s max_migration.txt contains"
                        + f" {max_migration_name!r}, but the latest migration"
                        + f" is {real_max_migration_name!r}."
                    ),
                    hint=(
                        "Edit max_migration.txt to contain"
                        + f" {real_max_migration_name!r} or rearrange the"
                        + " migrations into the correct order."
                    ),
                )
            )

    return errors



from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from typing import Any

from django.db.migrations.writer import MigrationWriter


@contextmanager
def spy_on_migration_writers() -> Generator[dict[str, str]]:
    written_migrations = {}

    orig_as_string = MigrationWriter.as_string

    def wrapped_as_string(self: MigrationWriter, *args: Any, **kwargs: Any) -> str:
        written_migrations[self.migration.app_label] = self.migration.name
        return orig_as_string(self, *args, **kwargs)

    MigrationWriter.as_string = wrapped_as_string  # type: ignore [method-assign]
    try:
        yield written_migrations
    finally:
        MigrationWriter.as_string = orig_as_string  # type: ignore [method-assign]



from __future__ import annotations

from typing import Any

from django.core.management.commands.makemigrations import Command as BaseCommand

from django_linear_migrations.apps import MigrationDetails, first_party_app_configs
from django_linear_migrations.management.commands import spy_on_migration_writers


class Command(BaseCommand):
    def handle(self, *app_labels: Any, **options: Any) -> None:
        with spy_on_migration_writers() as written_migrations:
            super().handle(*app_labels, **options)

        if options["dry_run"]:
            return

        first_party_app_labels = {
            app_config.label for app_config in first_party_app_configs()
        }

        for app_label, migration_name in written_migrations.items():
            if app_label not in first_party_app_labels:
                continue

            # Reload required in case of initial migration
            migration_details = MigrationDetails(app_label, do_reload=True)
            max_migration_txt = migration_details.dir / "max_migration.txt"
            max_migration_txt.write_text(f"{migration_name}\n")
